/* turn ascii input into UTF8 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define BUFLEN 5000

/* KOREAN */

#define KR_BASE 0xAC00
#define KR_CONS (21*28)
#define KR_VOWEL 28

/* rewriting rules, applied in order */
char *kr_rule[][2]={
  {"t'", "T"}, {"k'", "K"}, {"p'", "f"}, {"ch'", "CH"},
  {"r", "l"}, {"kh", "K"}, {"ng", "G"}, {"P", "f"}, {"ph", "f"},
  {"tch", "CH"}, {"tsh", "CH"}, {"tss", "chch"}, {"j", "ch"}, 
  {"ts", "ch"}, {"sh", "s"}, {"th", "T"}, 
  {"t", "d"}, {"b", "p"}, {"g", "k"}, {"j", "ch"},
  {"oe", "oi"}, {"ee", "yi"}, {"yi", "eui"},
  {"y", "i"}, {"eo", "@"}, {"eu", "E"}, {"o", "O"}, {"u", "U"},
  {"iw", "i/w"}, {"wa", "Oa"}, {"wae", "Oae"}, {"wO", "U@"}, {"w@", "U@"}, 
  {"we", "Ue"}, {"wi", "Ui"}, {"UO", "U@"},
  {NULL, NULL}};

char *kr_cons[19]={
"k", "kk", "n", "d", "dd", "l", "m", "p", "pp", "s", "ss", "", "ch", "chch",
"CH", "K", "T", "f", "h"};
 
char *kr_vowels[21]={
"a", "ae", "ia", "iae", "@", "e", "i@", "ie", "O", "Oa", "Oae", "Oi", "iO",
"U", "U@", "Ue", "Ui", "iU", "E", "Ei", "i"};

char *kr_endings[28]={
"", "k", "kk", "ks", "n", "nch", "nh", "d", "l", "lk", "lm", "lp", "ls", "lT",
"lf", "lh", "m", "p", "ps", "s", "ss", "G", "ch", "CH", "K", "T", "f", "h"};

char *kr_sep="-/| \n0123456789.,?!()";

/* for decoding */
char *unkr_cons[19]={
"k", "kk", "n", "t", "tt", "l", "m", "p", "pp", "s", "ss", "", "j", "jj",
"ch'", "k'", "t'", "p'", "h"};
 
char *unkr_vowels[21]={
"a", "ae", "ya", "yae", "eo", "e", "yeo", "ye", "o", "wa", "wae", "oe", "yo",
"u", "wo", "we", "wi", "yu", "eu", "yi", "i"};

char *unkr_endings[28]={
"", "k", "kk", "ks", "n", "nj", "nh", "t", "l", "lk", "lm", "lp", "ls", "lt'",
"lp'", "lh", "m", "p", "ps", "s", "ss", "ng", "j", "ch'", "k'", "t'", "p'", "h"};

/* RUSSIAN */

char *ru_rule[][2]={
 {"j", "zh"}, {"J", "Zh"}, {"^", "~"}, {"Yu", "Iu"}, {"yu", "iu"},
 {"Ya", "Ia"}, {"ya", "ia"},
 {NULL, NULL}};

char *ru_alphabet[96]={
"E`", "E\"", "", "G'", "", "", "", "", "", "", "", "", "K'", "I`", "U~", "",
"A", "B", "V", "G", "D", "E", "Zh", "Z", "I", "I~", "K", "L", "M", "N", "O", 
"P", "R", "S", "T", "U", "F", "Kh", "Ts", "Ch", "Sh", "Shch", "\"~", "Y",
"'~", "E~", "Iu", "Ia", 
"a", "b", "v", "g", "d", "e", "zh", "z", "i", "i~", "k", "l", "m", "n", "o",
"p", "r", "s", "t", "u", "f", "kh", "ts", "ch", "sh", "shch", "\"", "y",
"'", "e~", "iu", "ia",
"e`", "e\"", "", "g'", "", "", "", "", "", "", "", "", "k'", "i`", "u~", ""};

char *ru_sep="/| \n0123456789.,?!()-";

/* GREEK */

char *gr_rule[][2]={
 {"s ", "s' "}, {"s\n", "s'\n"},
 {NULL, NULL}};

char *gr_alphabet[96]={
"", "", "", "", "", "", "", "", "", "", "", "", "", "", "?", "",
"", "", "", "", "", "", "A'", ";", "E'", "E:'", "I'", "", "O'", "", "U'", "O:'",
"i\"'", "A", "B", "G", "D", "E", "Z", "E:", "Th", "I", "K", "L", "M", "N", "X", "O",
"P", "R", "", "S", "T", "U", "Ph", "Kh", "Ps", "O:", "I\"", "U\"", "a'", "e'", "e:'", "i'",
"u\"'", "a", "b", "g", "d", "e", "z", "e:", "th", "i", "k", "l", "m", "n", "x", "o",
"p", "r", "s'", "s", "t", "u", "ph", "kh", "ps", "o:", "i\"", "u\"", "o'", "u'", "o:'", ""
};

char *gr_ext_alphabet[256]={
"'a", "`a", "'a`", "`a`", "'a'", "`a'", "'a~", "`a~",
"'A", "`A", "'A`", "`A`", "'A'", "`A'", "'A~", "`A~",
"'e", "`e", "'e`", "`e`", "'e'", "`e'", "", "",
"'E", "`E", "'E`", "`E`", "'E'", "`E'", "", "",
"'e:", "`e:", "'e:`", "`e:`", "'e:'", "`e:'", "'e:~", "`e:~",
"'E:", "`E:", "'E:`", "`E:`", "'E:'", "`E:'", "'E:~", "`E:~",
"'i", "`i", "'i`", "`i`", "'i'", "`i'", "'i~", "`i~",
"'I", "`I", "'I`", "`I`", "'I'", "`I'", "'I~", "`I~",
"'o", "`o", "'o`", "`o`", "'o'", "`o'", "", "",
"'O", "`O", "'O`", "`O`", "'O'", "`O'", "", "",
"'u", "`u", "'u`", "`u`", "'u'", "`u'", "'u~", "`u~",
"", "`U", "", "`U`", "", "`U'", "", "`U~",
"'o:", "`o:", "'o:`", "`o:`", "'o:'", "`o:'", "'o:~", "`o:~",
"'O:", "`O:", "'O:`", "`O:`", "'O:'", "`O:'", "'O:~", "`O:~",
"a`", "a'", "e`", "e'", "e:`", "e:'", "i`", "i'",
"o`", "o'", "u`", "u'", "o:`", "o:'", "", "",
"'a_", "`a_", "'a_`", "`a_`", "'a_'", "`a_'", "'a_~", "`a_~",
"'A_", "`A_", "'A_`", "`A_`", "'A_'", "`A_'", "'A_~", "`A_~",
"'e:_", "`e:_", "'e:_`", "`e:_`", "'e:_'", "`e:_'", "'e:_~", "`e:_~",
"'E:_", "`E:_", "'E:_`", "`E:_`", "'E:_'", "`E:_'", "'E:_~", "`E:_~",
"'o:_", "`o:_", "'o:_`", "`o:_`", "'o:_'", "`o:_'", "'o:_~", "`o:_~",
"'O:_", "`O:_", "'O:_`", "`O:_`", "'O:_'", "`O:_'", "'O:_~", "`O:_~",
"a!", "a:", "a_`", "a_", "a_'", "", "a~", "a_~",
"A!", "A:", "A`", "A'", "A_", "", "", "'",
"", "", "e:_`", "e:_", "e:_'", "", "e:~", "e:_~",
"E`", "E'", "E:`", "E:'", "E:_", "'`", "''", "'~",
"i!", "i:", "i\"`", "i\"'", "", "", "i~", "i\"~", 
"I!", "I:", "I`", "I'", "", "``", "`'", "`~",
"u!", "u:", "u\"`", "u\"'", "'r", "`r", "u~", "u\"~",
"U!", "U:", "U`", "U'", "`R", "", "", "",
"", "", "o:_`", "o:_", "o:_'", "", "o:~", "o:_~",
"O`", "O'", "O:`", "O:'", "O:_", "", "`", ""};

char *gr_sep="/| \n0123456789.,!()-";

/* LATIN */

char *latin[0x1A0]={
" ", "/!", "/c", "=L", "xo", "=Y", "||", "/S",
"\"", "(c)", "^a", "<<", "-|", "", "(r)", "^-",
"^0", "+-", "^2", "^3", "'", "mu", "/P", ".", 
",", "^1", "^o", ">>", "1/4", "1/2", "3/4", "/?",
"A`", "A'", "A^", "A~", "A\"", "A^o", "AE", "C,", 
"E`", "E'", "E^", "E\"", "I`", "I'", "I^", "I\"",
"/D", "N~", "O`", "O'", "O^", "O~", "O\"", "x", 
"/O", "U`", "U'", "U^", "U\"", "Y'", "|P", "ss",
"a`", "a'", "a^", "a~", "a\"", "a^o", "ae", "c,",
"e`", "e'", "e^", "e\"", "i`", "i'", "i^", "i\"",
"/d", "n~", "o`", "o'", "o^", "o~", "o\"", "/", 
"/o", "u`", "u'", "u^", "u\"", "y'", "|p", "y\"",

"A^-", "a^-", "A^u", "a^u", "A,", "a,", "C'", "c'",
"C^", "c^", "C^.", "c^.", "C^v", "c^v", "D^v", "d^v",
"-D", "-d", "E^-", "e^-", "E^u", "e^u", "E^.", "e^.",
"E,", "e,", "E^v", "e^v", "G^", "g^", "G^u", "g^u",
"G^.", "g^.", "G,", "g,", "H^", "h^", "-H", "-h",
"I~", "i~", "I^-", "i^-", "I^u", "i^u", "I,", "i,",
"I^.", "\\i", "IJ", "ij", "J^", "j^", "K,", "k,", 
"_k", "L'", "l'", "L,", "l,", "L^v", "l^v", "L^.",
"l^.", "/L", "/l", "N'", "n'", "N,", "n,", "N^v", 
"n^v", "'n", "N_,", "n_,", "O^-", "o^-", "O^u", "o^u",
"O''", "o''", "OE", "oe", "R'", "r'", "R,", "r,",
"R^v", "r^v", "S'", "s'", "S^", "s^", "S,", "s,", 
"S^v", "s^v", "T,", "t,", "T^v", "t^v", "-T", "-t",
"U~", "u~", "U^-", "u^-", "U^u", "u^u", "U^o", "u^o",
"U''", "u''", "U,", "u,", "W^", "w^", "Y^", "y^", 
"Y\"", "Z'", "z'", "Z^.", "z^.", "Z^v", "z^v", "|s",

"-b", "B;", "B^-", "b^-", "", "", "", "C;",
"c;", "", "D;", "D^-", "d^-", "", "", "",
"", "F,", "f,", "G;", "", "hv", "", "-I",
"K;", "k;", "-l", "", "", "", "", "-O",
"O)", "o)", "OI", "oi", "P;", "p;", "", "",
"", "", "", "", "T;", "t;", "", "U)",
"u)", "", "V;", "", "", "-Z", "-z", "",
"", "", "", "", "", "", "", "wy",
"", "", "", "", "DZ^v", "Dz^v", "dz^v", "LJ",
"Lj", "lj", "NJ", "Nj", "nj", "A^v", "a^v", "I^v",
"i^v", "O^v", "o^v", "U^v", "u^v", "U\"^-", "u\"^-", "U\"'",
"u\"'", "U\"^v", "u\"^v", "U\"`", "u\"`", "", "A\"^-", "a\"^-",
"A^.^-", "a^.^-", "AE^-", "ae^-", "-G", "-g", "G^v", "g^v",
"K^v", "k^v", "O,", "o,", "O,^-", "o,^-", "", "",
"j^v", "DZ", "Dz", "dz", "G'", "g'", "HV", "WY",
"N`", "n`", "A^o'", "a^o'", "AE'", "ae'", "/O'", "/o'",
"A``", "a``", "A^n", "a^n", "E``", "e``", "E^n", "e^n",
"I``", "i``", "I^n", "i^n", "O``", "o``", "O^n", "o^n",
"R``", "r``", "R^n", "r^n", "U``", "u``", "U^n", "u^n",
"S_,", "s_,", "T_,", "t_,", "", "", "H^v", "h^v",
"", "", "", "", "Z;", "z;", "A^.", "a^.",
"E_,", "e_,", "O\"^-", "o\"^-", "O~^-", "o~^-", "O^.", "o^.",
"O^.^-", "o^.^-", "Y^-", "y^-", "", "", "", "",
"", "", "", "", "", "", "", "",
};

char *latin_ext[256]={
"A_o", "a_o", "B^.", "b^.", "B.", "b.", "B_", "b_", 
"C,'", "c,'", "D^.", "d^.", "D.", "d.", "D_", "d_", 
"D,", "d,", "D_^", "d_^", "E^-`", "e^-`", "E^-'", "e^-'",
"E_^", "e_^", "E_~", "e_~", "E,^u", "e,^u", "F^.", "f^.",
"G^-", "g^-", "H^.", "h^.", "H.", "h.", "H\"", "h\"",
"H,", "h,", "H_u", "h_u", "I_~", "i_~", "I\"'", "i\"'",
"K'", "k'", "K.", "k.", "K_", "k_", "L.", "l.", 
"L.^-", "l.^-", "L_", "l_", "L_^", "l_^", "M'", "m'",
"M^.", "m^.", "M.", "m.", "N^.", "n^.", "N.", "n.",
"N_", "n_", "N_^", "n_^", "O~'", "o~'", "O~\"", "o~\"",
"O^-`", "o^-`", "O^-'", "o^-'", "P'", "p'", "P^.", "p^.",
"R^.", "r^.", "R.", "r.", "R.^-", "r.^-", "R_", "r_",
"S^.", "s^.", "S.", "s.", "S'^.", "s'^.", "S^v.", "s^v.",
"S.^.", "s.^.", "T^.", "t^.", "T.", "t.", "T_", "t_",
"T_^", "t_^", "U_\"", "u_\"", "U_~", "u_~", "U_^", "u_^",
"U~'", "u~'", "U^-\"", "u^-\"", "V~", "v~", "V.", "v.", 
"W`", "w`", "W'", "w'", "W\"", "w\"", "W^.", "w^.", 
"W.", "w", "X^.", "x^.", "X\"", "x\"", "Y^.", "y^.",
"Z^", "z^", "Z.", "z.", "Z_", "z_", "h_", "t\"", 
"w^o", "y^o", "", "|s^.", "", "", "", "",
"A.", "a.", "A;", "a;", "A^'", "a^'", "A^`", "a^`", 
"A^;", "a^;", "A^~", "a^~", "A.^", "a.^", "A^u'", "a^u'",
"A^u`", "a^u`", "A^u;", "a^u;", "A^u~", "a^u~", "A.^u", "a.^u",
"E.", "e.", "E;", "e;", "E~", "e~", "E^'", "e^'", 
"E^`", "e^`", "E^;", "e^;", "E^~", "e^~", "E.^", "e.^", 
"I;", "i;", "I.", "i.", "O.", "o.", "O;", "o;",
"O^'", "o^'", "O^`", "o^`", "O^;", "o^;", "O^~", "o^~",
"O.^", "o.^", "O)'", "o)'", "O)`", "o)`", "O);", "o);",
"O)~", "o)~", "O.)", "o.)", "U.", "u.", "U;", "u;",
"U)'", "u)'", "U)`", "u)`", "U);", "u);", "U)~", "u)~",
"U.)", "u.)", "Y`", "y`", "Y.", "y.", "Y;", "y;", 
"Y~", "y~", "", "", "", "", "", ""
};

char buf1[BUFLEN], buf2[BUFLEN];
int curbuf;

/* decode a korean character */
int attempt_hangul(char **ptr, int validateonly)
{
  int unicode, match, matchlen, i;
  char *p, *q;
  
  p = *ptr;
  unicode = KR_BASE;
  
  /* first a consonant */
  match = 11; matchlen = 0;
  for (i=0; i<19; i++)
    if ((strlen(kr_cons[i])>=matchlen+1) && 
        !strncmp(p, kr_cons[i], strlen(kr_cons[i])))
      { match = i; matchlen = strlen(kr_cons[i]); }
  p+=matchlen;
  unicode += match * KR_CONS;
  
  /* then a vowel */
  match = -1; matchlen = -1;
  for (i=0; i<21; i++)
    if ((strlen(kr_vowels[i])>=matchlen+1) && 
        !strncmp(p, kr_vowels[i], strlen(kr_vowels[i])))
      { match = i; matchlen = strlen(kr_vowels[i]); }
  if (match<0)
    return -1;
  p+=matchlen;
  unicode += match * KR_VOWEL;
  
  /* finally an optional consonant, but don't be greedy */
  if (validateonly) return unicode;
  match = 0; matchlen = 0;
  q=p; 
  if (attempt_hangul(&q, 1)<0) /* do we need to include a final */
    for (i=0; i<28; i++)
      if ((strlen(kr_endings[i])>=matchlen+1) && 
          !strncmp(p, kr_endings[i], strlen(kr_endings[i])))
        { 
          if (i==3 || i==5 || i==6 || (i>=9 && i<=15) || i==18)
            { q = p+strlen(kr_endings[i]);
              if (*q=='a' || *q=='e' || *q=='i' || *q=='0' || *q=='O' || 
                  *q=='E' || *q=='U') continue;  /* too greedy */
            }
          match = i; matchlen = strlen(kr_endings[i]); 
        }
  p+=matchlen;
  unicode += match;
  *ptr = p;
  return unicode;
}

/* output a unicode character in utf8 */
void uniput(int unicode)
{
  if (unicode<0 || unicode>0xFFFF) putchar('?');
  else if (unicode<0x80) putchar(unicode);
  else if (unicode<0x800) {
    putchar(0xC0 + (unicode>>6));
    putchar(0x80 + (unicode&0x3F));
  } else {
    putchar(0xE0+(unicode>>12));
    putchar(0x80+((unicode>>6)&0x3F));
    putchar(0x80+(unicode&0x3F));
  }
}

/* utf8 to unicode */
int uniget(char **p)
{
  unsigned char *q;
  int code;
  
  q=(unsigned char *)(*p);
  if (*q<0x80) { (*p)++; return (int)*q; }
  else if (*q<0xc0) { (*p)++; return -1; }
  else if (*q<0xe0) {
    (*p)+=2;
    code = ((*q)-0xc0)<<6;
    if (q[1]<0x80 || q[1]>=0xc0) return -1;
    code += (q[1] - 0x80);
    return code;
  }
  else if (*q<0xf0) {
    (*p)+=3;
    code = ((*q)-0xe0)<<12;
    if (q[1]<0x80 || q[1]>=0xc0) return -1;
    code += (q[1]-0x80)<<6;
    if (q[2]<0x80 || q[2]>=0xc0) return -1;
    code += (q[2]-0x80);
    return code;
  }
  else if (*q<0xf8) { (*p)+=4; return -1; }
  else if (*q<0xfc) { (*p)+=5; return -1; }
  else if (*q<0xfe) { (*p)+=6; return -1; }
  else { (*p)++; return -1; }
}

/* rewrite a line in buf1 according to certain rules applied in sequence */
void do_rewrite(char *rule[][2])
{
  int i;
  char *p, *q;
  
  for (i = 0; rule[i][0]!=NULL; i++) {
    if (i&1) { p=buf2; q=buf1;} else { p=buf1; q=buf2; }
    while (*p!=0) {
      if (!strncmp(p, rule[i][0], strlen(rule[i][0]))) {
        p+=strlen(rule[i][0]);
        strcpy(q,rule[i][1]);
        q+=strlen(rule[i][1]);
      } else *(q++)=*(p++);
    }
    *q=0;
  }
  if (i&1) strcpy(buf1, buf2);
}

int main(int argc, char **argv)
{
  int i, unicode, match, matchlen;
  char *p;

  if (argc!=2 || argv[1][0]!='-' || argv[1][1]=='h') {
    fprintf(stderr,
           "mkutf8 v0.1 by Denis Auroux (auroux@math.polytechnique.fr)\n"
           "Usage: mkutf8 [-l|-g|-r|-k] < ascii-file > utf8-file\n"
           "       mkutf8 -d < utf8-file > ascii-file\n"
           "    (-l: latin, -g: greek, -r: russian, -k: korean)\n");
    exit(1);
  }

  switch(argv[1][1]) {
    case 'k': /* KOREAN */
      while (1) {
        fgets(buf1, BUFLEN/2, stdin);
        if (feof(stdin)) exit(0);
        /* first rewrite the line according to phonetic rewriting rules */
        do_rewrite(kr_rule);
        /* then do the hangul translation */
        p=buf1;
        while (*p!=0) {
          while (*p!=0 && strchr(kr_sep, *p)!=NULL)
            { if (*p!='-' && *p!='/' && *p!='|') putchar(*p);   p++; }
          if (*p==0) break;
          unicode = attempt_hangul(&p,0);
          if (unicode < 0) { putchar('?'); p++; }
          else uniput(unicode);
        }
      }
      break;
  
    case 'r': /* RUSSIAN */
      while (1) {
        fgets(buf1, BUFLEN/2, stdin);
        if (feof(stdin)) exit(0);
        /* first rewrite the line according to phonetic rewriting rules */
        do_rewrite(ru_rule);
        /* then do the russian translation */
        p=buf1;
        while (*p!=0) {
          while (*p!=0 && strchr(ru_sep, *p)!=NULL)
            { if (*p!='/' && *p!='|') putchar(*p);   p++; }
          if (*p==0) break;
          match = -1; matchlen = 0;
          for (i=0; i<96; i++)
            if ((strlen(ru_alphabet[i])>=matchlen+1) && 
                !strncmp(p, ru_alphabet[i], strlen(ru_alphabet[i])))
              { match = i; matchlen = strlen(ru_alphabet[i]); }
          if (match<0) { putchar('?'); p++; }
          else { 
            p+=matchlen; 
            uniput(0x400+match);
          }
        }
      }
      break;

    case 'g': /* GREEK */
      while (1) {
        fgets(buf1, BUFLEN/2, stdin);
        if (feof(stdin)) exit(0);
        /* first rewrite the line according to phonetic rewriting rules */
        do_rewrite(gr_rule);
        /* then do the greek translation */
        p=buf1;
        while (*p!=0) {
          while (*p!=0 && strchr(gr_sep, *p)!=NULL)
            { if (*p!='/' && *p!='|') putchar(*p);   p++; }
          if (*p==0) break;
          match = -1; matchlen = 0;
          for (i=0; i<96; i++)
            if ((strlen(gr_alphabet[i])>=matchlen+1) && 
                !strncmp(p, gr_alphabet[i], strlen(gr_alphabet[i])))
              { match = i; matchlen = strlen(gr_alphabet[i]); }
          if (match<0) unicode = -1; else unicode = 0x370 + match;
          match = -1;
          for (i=0; i<256; i++)
            if ((strlen(gr_ext_alphabet[i])>=matchlen+1) && 
                !strncmp(p, gr_ext_alphabet[i], strlen(gr_ext_alphabet[i])))
              { match = i; matchlen = strlen(gr_ext_alphabet[i]); }
          if (match>=0) unicode = 0x1F00 + match;
          if (unicode<0) { putchar('?'); p++; }
          else { 
            p+=matchlen; 
            uniput(unicode);
          }
        }
      }
      break;

    case 'l': /* LATIN */
      while (1) {
        fgets(buf1, BUFLEN, stdin);
        if (feof(stdin)) exit(0);
        /* no rewriting; sequences are {}-delimited */
        p=buf1;
        while (*p!=0) {
          if (*p!='{') {
            if ((*p>=0x20 && *p<=0x7E) || *p=='\n' || *p=='\t') 
              putchar(*p); else putchar('?');
            p++; continue;
          }
          p++;
          for (matchlen=0; p[matchlen]!=0; matchlen++)
            if (p[matchlen]=='}') break;
          if (p[matchlen]!='}') { putchar('{'); continue; }
          if (matchlen==0) { p++; continue; }
          unicode = -1;
          if (matchlen==1 && p[0]>=0x20 && p[0]<=0x7E) unicode = p[0];
          if (matchlen==2) {
            /* euro */
            if ((p[0]=='=' && p[1]=='E') || (p[0]=='E' && p[1]=='=')) 
              unicode = 0x20AC;
            /* other reversed length 2 sequences */
            for (i=0; i<256; i++)
              if (p[1]==latin_ext[i][0] && p[0]==latin_ext[i][1] && !latin_ext[i][2])
                unicode = 0x1E00 + i;
            for (i=0; i<0x1A0; i++)
              if (p[1]==latin[i][0] && p[0]==latin[i][1] && !latin[i][2])
                unicode = 0xA0 + i;
          }
          /* normal sequences */
          for (i=0; i<256; i++) 
            if (strlen(latin_ext[i])==matchlen && !strncmp(p, latin_ext[i], matchlen))
              unicode = 0x1E00 + i;
          for (i=0; i<0x1A0; i++)
            if (strlen(latin[i])==matchlen && !strncmp(p, latin[i], matchlen))
              unicode = 0xA0 + i;
          p += matchlen+1;
          if (unicode<0) putchar('?'); else uniput(unicode);
        }
      }
      break;

    case 'd': /* DECODING */
      while (1) {
        fgets(buf1, BUFLEN-10, stdin);
        if (feof(stdin)) exit(0);
        p = buf1+strlen(buf1);
        for (i=1;i<=6;i++) p[i] = 0; /* avoid skipping past end */
        p = buf1;
        while (*p!=0) {
          unicode = uniget(&p);
          if (unicode < 0) putchar('?');
          else if (unicode == 0) break;
          else if (unicode < 0x80) putchar(unicode); /* ascii */
          else if (unicode >=0x400 && unicode < 0x460) { /* russian */
            if (ru_alphabet[unicode-0x400][0]) 
              printf("%s", ru_alphabet[unicode-0x400]);
            else putchar('?');
          } 
          else if (unicode >=0x370 && unicode < 0x3D0) { /* basic greek */
            if (gr_alphabet[unicode-0x370][0])
              printf("%s", gr_alphabet[unicode-0x370]);
            else putchar('?');
          } 
          else if (unicode >=0x1F00 && unicode < 0x2000) { /* ext'd greek */
            if (gr_ext_alphabet[unicode-0x1F00][0])
              printf("%s", gr_ext_alphabet[unicode-0x1F00]);
            else putchar('?');
          } 
          else if (unicode >=KR_BASE && unicode < KR_BASE + 19*21*28) { /* hangul */
            printf("%s%s%s", unkr_cons[(unicode - KR_BASE)/KR_CONS],
               unkr_vowels[((unicode - KR_BASE)%KR_CONS)/KR_VOWEL],
               unkr_endings[(unicode - KR_BASE)%KR_VOWEL]);
          }
          else if (unicode==0x20AC) /* euro */
            printf("{=E}");
          else if (unicode >= 0xA0 && unicode < 0x240) { /* latin */
            if (latin[unicode-0xA0][0])
              printf("{%s}", latin[unicode-0xA0]);
            else putchar('?');
          }
          else if (unicode >= 0x1E00 && unicode < 0x1F00) { /* latin++ */
            if (latin_ext[unicode-0x1E00][0])
              printf("{%s}", latin_ext[unicode-0x1E00]);
            else putchar('?');
          }
          else putchar('?');
        }
      }
      break;
  
    default:
      fprintf(stderr, "Unsupported translation !\n");
      exit(1);
  }
}